import pytest

import slotted


class Foo:
    __slots__ = ("foo",)


class Bar(Foo):
    __slots__ = ("bar",)


class ForcedBarM(Bar, metaclass=slotted.SlottedMeta):
    pass


class ForcedBar(Bar, slotted.Slotted):
    pass


class ForcedBarABCM(ForcedBarM, metaclass=slotted.SlottedABCMeta):
    pass


class ForcedBarABC(ForcedBar, slotted.SlottedABC):
    pass


def test_slots():
    assert slotted.slots(Foo) == {"foo"}
    assert slotted.slots(Bar) == {"foo", "bar"}
    for forced_cls in (ForcedBarM, ForcedBar, ForcedBarABCM, ForcedBarABC):
        assert hasattr(forced_cls, "__slots__")
        assert slotted.slots(forced_cls) == {"foo", "bar"}


def test_non_slotted():
    class NonSlotted:
        pass

    with pytest.raises(TypeError):
        slotted.SlottedMeta("NonSlotted", (NonSlotted,), {})

    with pytest.raises(TypeError):
        slotted.SlottedABCMeta("NonSlotted", (NonSlotted,), {})

    with pytest.raises(TypeError):
        type("NonSlotted", (NonSlotted, slotted.Slotted), {})

    with pytest.raises(TypeError):
        type("NonSlotted", (NonSlotted, slotted.SlottedABC), {})


if __name__ == "__main__":
    pytest.main()



import abc
import inspect
from typing import Set, Type

__all__ = ["SlottedMeta", "Slotted", "SlottedABCMeta", "SlottedABC", "slots"]


class SlottedMeta(type):
    """Metaclass that enforces __slots__."""

    @staticmethod
    def __new__(mcs, name, bases, dct, **kwargs):

        # Force slots when constructing this class.
        if "__slots__" not in dct:
            dct = dict(dct)
            dct["__slots__"] = ()
        cls = super().__new__(mcs, name, bases, dct, **kwargs)

        # All bases are required to have slots.
        for base in inspect.getmro(cls)[1:-1]:
            if not hasattr(base, "__slots__"):
                raise TypeError(f"{base} does not define __slots__")

        return cls


class Slotted(metaclass=SlottedMeta):
    """Enforces __slots__."""

    __slots__ = ()


class SlottedABCMeta(abc.ABCMeta, SlottedMeta):
    """Mix-in of abc.ABCMeta and SlottedMeta."""


class SlottedABC(abc.ABC, Slotted, metaclass=SlottedABCMeta):
    """Mix-in of abc.ABC and Slotted."""

    __slots__ = ()


def slots(cls: Type) -> Set[str]:
    """Get all slot names for a class."""
    _slots: Set[str] = set()
    for base in inspect.getmro(cls)[:-1]:
        _slots.update(getattr(base, "__slots__", ()))
    return _slots



import setuptools

with open("README.rst", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="slotted",
    version="2.0.0",
    author="Bruno Nicko",
    author_email="brunonicko@gmail.com",
    description="Enforces usage of '__slots__' for python classes",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/brunonicko/slotted",
    py_modules=["slotted"],
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
    python_requires=">=3.7",
    tests_require=["pytest"],
)
